from voca import utils
from voca import context
from voca.plugins import basic


registry = utils.Registry()
registry.define({"?text": r"/\w.+/"})


@registry.register('"hit" chord')
async def _say_def(message):
    await basic.write("def")


wrapper = utils.Wrapper(registry)



from __future__ import annotations

import subprocess

from typing_extensions import Protocol

import attr
import trio


from voca import platforms
from voca import log


@platforms.implementation(platforms.System.WINDOWS, platforms.System.DARWIN)
async def get_current_window_title():
    import pygetwindow

    window = await trio.run_sync_in_worker_thread(pygetwindow.getFocusedWindow)
    return window.title


@platforms.implementation(platforms.System.LINUX)
async def get_current_window_title():
    proc = await utils.run_subprocess(
        ["/usr/bin/xdotool", "getwindowfocus", "getwindowname"], stdout=subprocess.PIPE
    )
    return proc.stdout.decode()[:-1]


@attr.dataclass
class WindowContext:
    title: str

    async def check(self, data=None) -> bool:
        current_title = await get_current_window_title()
        return self.title in current_title


@attr.dataclass
class AlwaysContext:
    async def check(self, data=None) -> bool:
        return True


@attr.dataclass
class NeverContext:
    async def check(self, data=None) -> bool:
        return False


@log.log_async_call
async def filter_wrappers(
    wrapper_group: utils.WrapperGroup, data: dict
) -> utils.WrapperGroup:
    allowed = []
    for wrapper in wrapper_group.wrappers:

        if await wrapper.context.check(data):
            allowed.append(wrapper)
    return utils.WrapperGroup(allowed)


from voca import utils



from __future__ import annotations

import contextlib
import sys
import types
import re

from typing import Optional
from typing import List
from typing_extensions import Protocol


import attr
import trio
import lark

import voca.plugins
from voca import utils
from voca import context
from voca import log
from voca import patching


class LazyLoader:
    def __getattr__(self, name):
        from voca.plugins import basic

        return getattr(basic, name)


basic = LazyLoader()


class AsyncActionType(Protocol):
    async def execute(self, arg=None) -> None:
        ...


@attr.dataclass
class TextAction:
    text: str

    async def execute(self, arg=None) -> None:
        await basic.write(self.text.format_map(arg))


@attr.dataclass
class KeyAction:
    name: str

    async def execute(self, arg=None):
        await basic.press(self.name)


@attr.dataclass
class RegisteredAction:
    instruction: AsyncActionType = attr.ib()
    rdescript: Optional[str] = attr.ib(default=None)

    async def execute(self, arg=None) -> None:
        await self.instruction.execute(arg)


@attr.dataclass
class ActionSequence:
    actions: List[AsyncActionType]

    def __add__(self, other):
        return attr.evolve(self, actions=self.actions + other.actions)

    async def execute(self, arg=None):

        for action in self.actions:
            await action.execute(arg)


@attr.dataclass
class ConditionalAction:
    condition: types.FunctionType
    action: RegisteredAction

    async def execute(self, arg=None):
        if await self.condition(arg):
            await self.action(arg)


# @attr.dataclass
# class AppContext:
#     title: str
#     executable: str


@attr.dataclass
class Dictation:
    name: str

    def make_definitions(self):
        return {self.name: r"/\w+/"}


@attr.dataclass
class IntegerRefST:
    name: str
    start: int
    end: int

    def make_definitions(self):
        mapping = utils.value_to_pronunciation()
        alternatives = [
            mapping.get(str(i)) for i in range(self.start, min(self.end, 20))
        ]
        return {self.name: "/" + "|".join(alternatives) + "/"}


@attr.dataclass
class RepeatedAction:
    action: AsyncActionType
    extra: str

    async def execute(self, arg=None):

        # self.extra should be the name of the extra, like 'n'
        # arg should be a mapping of values said
        times = int(arg[self.extra])
        for _ in range(times):
            await self.action.execute(arg)


@attr.dataclass
class DelayAction:
    duration: float

    async def execute(self, arg=None):
        await trio.sleep(self.duration)


@attr.dataclass
class Choice:
    name: str
    mapping: dict

    def make_definitions(self):
        definitions = {k: utils.quote(v) for k, v in self.mapping.items()}
        definitions[self.name] = "|".join(definitions.keys())
        return definitions


@attr.dataclass
class MyRepeat:
    extra: str

    def __mul__(self, action):
        return RepeatedAction(action, self.extra)

    def __rmul__(self, action):
        return self * action


@attr.dataclass
class FunctionAction:
    callable: types.FunctionType
    positional_arguments: tuple
    keyword_arguments: dict

    async def execute(self, arg):
        return self.callable(*self.positional_arguments, **self.keyword_arguments)


class CasterTransformer(lark.Transformer):
    def n(self, arg):
        pronunciation = "".join(arg[0])
        value = utils.pronunciation_to_value()[pronunciation]
        return {"n": int(value)}


def transform_tree(tree):
    if not tree:
        return tree
    result = CasterTransformer().transform(tree[0])
    return result


async def _run(action, data):
    result = transform_tree(data)
    await action.execute(result)


def add_to_registry(mapping, registry):
    # TODO don't mutate registry
    for pattern, action in mapping.items():
        registry.register(utils.quote(pattern))(lambda data: _run(action, data))

    return registry


def _function(f, **kwargs):
    return FunctionAction(f, positional_arguments=(), keyword_arguments=kwargs)


class MergeRule:
    pass


class CCRMerger:
    CORE = 1


@attr.dataclass
class Grammar:
    name: str
    context: AppContext


class Settings:
    def __getitem__(self, name):
        return Settings()

    def __bool__(self):
        return False


class SpecTransformer(lark.Transformer):
    def name(self, args):
        return "".join(args)

    def literal_name(self, args):
        return " " + utils.quote(args[0].strip()) + " "

    def angled_name(self, args):
        return args[0]

    def optional_component(self, args):
        return f"[{args[0]}]"

    def phrase_component(self, args):
        return args[0]

    def component(self, args):
        return args[0]

    def spec(self, args):
        return "".join(args).strip()

    def group(self, args):
        return "(" + "|".join(args) + ")"


def convert_spec(spec):

    pattern = r"""

    ?start : spec
    spec : component+
    component : optional_component | phrase_component | group
    optional_component : "[" component "]"
    group : "(" component ("|" component)+ ")"
    phrase_component : angled_name | literal_name
    angled_name : "<" name ">"
    literal_name : /[\w ]+/
    name : /\w+/


    %import common.WS

    %ignore WS

    """

    parser = lark.Lark(pattern, debug=True, maybe_placeholders=True)
    tree = parser.parse(spec)
    result = "".join(SpecTransformer().transform(tree))
    result = result.replace('""', "")
    # XXX
    # if result.startswith("[") and result.endswith("]"):
    #     raise NotImplementedError(result + ": Buggy translator")
    return result


def adapt_AppContext(title=None, executable=None):
    return context.WindowContext(title=title)


def adapt_Dictation(text):
    return Dictation(text)


def convert_key_name(name):
    modifiers, _dash, simple = name.rpartition("-")
    modifier_map = {"c": "ctrl_l", "a": "alt", "s": "shift", "w": "cmd"}
    key_map = {"pgup": "page_up", "pgdown": "page_down"}
    new_modifiers = [utils.KeyModifier(modifier_map[m]) for m in modifiers]
    return utils.KeyChord(new_modifiers, key_map.get(simple, simple))


def adapt_Key(text):

    keys = [key.strip() for key in text.split(",")]
    actions = []
    for key in keys:
        name, _slash, delay = key.partition("/")

        actions.append(KeyAction(convert_key_name(name)))
        if delay:
            actions.append(DelayAction(float(delay)))

    return ActionSequence(actions)


def adapt_Text(name):
    return ActionSequence([TextAction(name)])


def find_merge_rule_classes(module):
    rules = []
    for obj in module.__dict__.values():

        if (
            isinstance(obj, type)
            and issubclass(obj, Placeholder)
            and obj is not Placeholder
        ):
            rules.append(obj)

    return rules


@attr.dataclass
class Patch:
    module: types.ModuleType
    name: str
    new: object


@contextlib.contextmanager
def monkeypatch(module, name, new):
    original = getattr(module, name)
    setattr(module, name, new)
    yield
    setattr(module, name, original)


def monkeypatch_each(patches):
    for patch in patches:
        monkeypatch(patch.module, patch.name, patch.new)


@log.log_call
def add_wrapper(module):
    if getattr(module, "wrapper", False):
        return module

    rules = find_merge_rule_classes(module)

    registry = utils.Registry()

    for rule in rules:
        for spec, action in rule.mapping.items():
            try:

                converted_spec = convert_spec(spec)
            except (lark.exceptions.UnexpectedCharacters, NotImplementedError):
                # XXX
                converted_spec = '"aklsdjfkldjfk"'
            # TODO Make this easier to read.
            registry.register(converted_spec)(
                lambda data, action=action: _run(action, data)
            )
        for extra in rule.extras:
            registry.define(extra.make_definitions())

    # TODO use module.context
    wrapper = utils.Wrapper(registry, context=context.AlwaysContext())
    module.wrapper = wrapper
    # XXX Avoid mutate-and-return.
    return module


def patch_all():

    # XXX Should fix this to avoid having to actually import castervoice.
    # Currently sys.modules is used to find castervoice in order to find its submodules.
    import castervoice

    finder = patching.make_finder(module_mapping)
    sys.meta_path.insert(0, finder)

    # XXX Maybe replace this with an import hook.
    utils.MODULE_TRANSFORMERS.append(add_wrapper)


class AttributeHaver:
    def __getattr__(self, name):
        return AttributeHaver()


class Placeholder:
    pass


class VirtualModule:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class VirtualPackage:
    def __init__(self, path, contents):
        self.__path__ = path
        self.__dict__.update(contents)


module_mapping = {
    "dragonfly.__init__": {},
    "dragonfly": {
        "Grammar": Grammar,
        "Context": None,
        "AppContext": adapt_AppContext,
        "Dictation": adapt_Dictation,
        "Repeat": MyRepeat,
        "Function": _function,
        "Choice": Choice,
        "Mouse": None,
        "Pause": None,
    },
    # "castervoice.lib": {},
    "castervoice.lib.control": {},
    "castervoice.lib.utilities": {"simple_log": lambda: None},
    # "castervoice.apps": {},
    # "castervoice.apps.__init__": {},
    "castervoice.lib.settings": {"SETTINGS": Settings()},
    "castervoice.lib.actions": {"Key": adapt_Key, "Text": adapt_Text},
    "castervoice.lib.context": {"AppContext": adapt_AppContext},
    "castervoice.lib.dfplus.additions": {"IntegerRefST": IntegerRefST},
    "castervoice.lib.dfplus.merge": {"gfilter": None},
    "castervoice.lib.dfplus.merge.mergerule": {"MergeRule": Placeholder},
    "castervoice.lib.dfplus.state.short": {"R": RegisteredAction},
    "castervoice.lib.dfplus.merge.ccrmerger": {"CCRMerger": AttributeHaver()},
}



import contextlib
import os
import json
import secrets
import subprocess
import string

from click.testing import CliRunner
import pytest

from voca import cli

from tests import helpers


def test_main():
    runner = CliRunner()
    result = runner.invoke(cli.cli, [])

    assert result.output.startswith("Usage")
    assert result.exit_code == 0


def make_command(utterance, final=True):
    return {
        "status": 0,
        "segment": 0,
        "result": {"hypotheses": [{"transcript": utterance}], "final": final},
        "id": "eec37b79-f55e-4bf8-9afe-01f278902599",
    }


@pytest.mark.usefixtures("virtual_display")
def test_strict():
    """Strict mode, not eager mode, executes final commands."""
    utterances = ["say alpha", "say bravo", "mode", "say charlie", "say delta"]

    rows = [make_command(utterance, final=True) for utterance in utterances]
    lines = ("\n".join(json.dumps(row) for row in rows) + "\n").encode()

    with helpers.capture_keypresses() as typed:
        helpers.run(["manage", "-i", "voca.plugins.basic"], input=lines)

    expected = ["KEY_A", "KEY_B"]
    assert typed == expected


@pytest.mark.usefixtures("virtual_display")
def test_eager():
    """Eager mode, not strict mode, executes non-final commands."""
    utterances = ["say alpha", "say bravo", "mode", "say charlie", "say delta"]

    rows = [make_command(utterance, final=False) for utterance in utterances]
    lines = ("\n".join(json.dumps(row) for row in rows) + "\n").encode()

    with helpers.capture_keypresses() as typed:
        helpers.run(["manage", "-i", "voca.plugins.basic"], input=lines)

    expected = ["KEY_C", "KEY_D"]
    assert typed == expected


@pytest.mark.usefixtures("virtual_display", "turtle_window")
def test_context_always():

    utterances = ["hit alpha", "hit bravo"]

    rows = [make_command(utterance, final=True) for utterance in utterances]
    lines = ("\n".join(json.dumps(row) for row in rows) + "\n").encode()

    with helpers.capture_keypresses() as typed:
        helpers.run(
            [
                "manage",
                "-i",
                "voca.plugins.basic",
                "-i",
                "voca.plugins.yes",
                "-i",
                "voca.plugins.no",
            ],
            input=lines,
        )

    expected = ["KEY_D", "KEY_E", "KEY_F"] * 2
    assert typed == expected


@pytest.mark.usefixtures("virtual_display", "turtle_window")
def test_context_never():

    utterances = ["nope", "nope"]

    rows = [make_command(utterance, final=True) for utterance in utterances]
    lines = ("\n".join(json.dumps(row) for row in rows) + "\n").encode()

    with helpers.capture_keypresses() as typed:
        helpers.run(["manage"], input=lines)

    expected = []
    assert typed == expected


@pytest.mark.usefixtures("virtual_display", "turtle_window")
def test_app_context_matches():
    """The spec is matched and the action is executed if the app is open."""

    utterances = ["check", "check"]

    rows = [make_command(utterance, final=True) for utterance in utterances]
    lines = ("\n".join(json.dumps(row) for row in rows) + "\n").encode()

    with helpers.capture_keypresses() as typed:
        helpers.run(
            [
                "manage",
                "-i",
                "voca.plugins.basic",
                "-i",
                "voca.plugins.turtle_context",
            ],
            input=lines,
        )

    expected = ["KEY_T", "KEY_U", "KEY_R", "KEY_T"] * 2
    assert typed == expected


@pytest.mark.usefixtures("virtual_display", "idle_window")
def test_app_context_does_not_match():
    """The spec is matched and the action is executed if the app is open."""

    utterances = ["check", "check"]

    rows = [make_command(utterance, final=True) for utterance in utterances]
    lines = ("\n".join(json.dumps(row) for row in rows) + "\n").encode()

    with helpers.capture_keypresses() as typed:
        helpers.run(["manage"], input=lines)

    expected = []
    assert typed == expected



from __future__ import annotations

import functools
import os
import itertools
import sys
import pathlib
import subprocess
import json
import contextlib
import copy

from typing import List
from typing import Optional

import attr
import trio
import eliot


from voca import plugins
from voca import utils
from voca import streaming
from voca import log


def worker_cli(should_log, module_names: Optional[List[str]] = None) -> List[str]:
    if module_names is None:
        module_names = utils.plugin_module_paths()

    log_arg = "--log" if should_log else "--no-log"
    prefix = [sys.executable, "-m", "voca", log_arg, "worker"]
    command = prefix.copy()
    for module_name in module_names:
        command += ["-i", module_name]
    return command


@log.log_async_call
async def replay_child_messages(child: trio.Process) -> None:
    async for message_from_child in streaming.TerminatedFrameReceiver(
        child.stdout, b"\n"
    ):
        print(message_from_child.decode())


@log.log_call
def set_state(data, state):
    state = copy.deepcopy(state)
    body = data["result"]["hypotheses"][0]["transcript"]
    command, _space, _args = body.partition(" ")
    if command == "mode":
        state["modes"]["strict"] = not state["modes"]["strict"]
        return state

    return None


@log.log_async_call
async def delegate_task(data, worker, state, action):
    wrapped_data = dict(
        **data, state=state, eliot_task_id=action.serialize_task_id().decode()
    )
    await worker.stdin.send_all(json.dumps(wrapped_data).encode() + b"\n")


@attr.s
class Pool:
    num_workers: int = attr.ib(default=1)
    should_log: bool = attr.ib(default=True)
    module_names: list = attr.ib(factory=list)
    processes: set = attr.ib(factory=set)

    def start(self):
        for _ in range(self.num_workers):
            self.add_new_process()

    def get_process(self):
        return self.processes.pop()

    def add_new_process(self):
        self.processes.add(
            trio.Process(
                worker_cli(self.should_log, self.module_names),
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
            )
        )


@log.log_async_call
async def run_worker(data, state, pool):

    with eliot.start_action(action_type="run_with_work") as action:
        worker = pool.get_process()

        await delegate_task(data=data, state=state, worker=worker, action=action)
        await replay_child_messages(worker)

        await worker.wait()
        pool.add_new_process()


@log.log_async_call
async def process_stream(receiver, num_workers, should_log, module_names):

    state = {"modes": {"strict": True}}

    pool = Pool(num_workers, should_log=should_log, module_names=module_names)
    pool.start()

    async for message_bytes in receiver:
        message = message_bytes.decode()
        with eliot.start_action(state=state):
            data = json.loads(message)
            if "result" not in data.keys():
                # Received a log, not a command.
                print(message)
                continue
            # Handle state changes.
            maybe_new_state = set_state(data, state)
            if maybe_new_state is not None:
                state = maybe_new_state
                continue

            # This logic could be moved into worker/plugin to allow for more modes.
            if not data["result"]["final"] and state["modes"]["strict"]:
                continue
            if data["result"]["final"] and not state["modes"]["strict"]:
                continue

            await run_worker(data=data, state=state, pool=pool)


@log.log_async_call
async def async_main(should_log, module_names: Optional[List[str]], num_workers: int):

    stream = trio._unix_pipes.PipeReceiveStream(os.dup(0))
    receiver = streaming.TerminatedFrameReceiver(stream, b"\n")

    await process_stream(
        receiver,
        num_workers=num_workers,
        should_log=should_log,
        module_names=module_names,
    )


@log.log_call
def main(should_log, module_names: Optional[List[str]], num_workers: int):
    trio.run(functools.partial(async_main, should_log, module_names, num_workers))



from __future__ import annotations

import contextlib
import collections
import sys
import types
import importlib
import importlib.machinery
import importlib.util

from typing import Optional
from typing import Dict
from typing import List
from typing import Any

import attr


@attr.s(auto_attribs=True)
class PathLoader:
    namespace: dict
    fullname: str
    path: Optional[str]
    target: types.ModuleType

    def create_module(self, spec: importlib.machinery.ModuleSpec):
        module = types.ModuleType(spec.name)
        module.__dict__.update(self.namespace)

        return module

    def exec_module(self, module):
        pass


@contextlib.contextmanager
def skipping_finder(finder):
    def placeholder_find_spec(_fullname, _path, _target):
        return None

    finder_placeholder = types.SimpleNamespace(find_spec=placeholder_find_spec)

    sys.meta_path[sys.meta_path.index(finder)] = finder_placeholder
    try:
        yield
    finally:
        sys.meta_path[sys.meta_path.index(finder_placeholder)] = finder


@contextlib.contextmanager
def skipping_module_in_sys_modules(module_name):

    # Use a sentinel to check if we need to restore parent module to sys.modules.
    sentinel = object()
    parent_module = sys.modules.pop(module_name, sentinel)
    try:
        yield
    finally:
        # Restore parent if necessary.
        if parent_module is not sentinel:

            # XXX Not sure if this is the right choice, or if it should leave
            # the existing entry in place.

            # Restore original module to sys.modules, bumping any entry created
            # during the `yield` above.
            sys.modules[module_name] = parent_module


@contextlib.contextmanager
def finder_patch(finder):
    sys.meta_path.insert(0, finder)
    try:
        yield
    finally:
        sys.meta_path.remove(finder)


@attr.s(auto_attribs=True)
class PathFinder:
    modules_to_handle: List[str]
    fullname_to_vars: Dict[str, Dict[str, Any]]

    def find_spec(
        self, fullname: str, path: Optional[str], target=Optional[types.ModuleType]
    ):

        if fullname not in self.modules_to_handle:
            # Return a spec using the default behavior ignoring sys.modules and
            # this finder.
            parent_name = ".".join(fullname.split(".")[:-1])

            with skipping_finder(self), skipping_module_in_sys_modules(parent_name):

                return importlib.util.find_spec(fullname, path)

        variables = self.fullname_to_vars.get(fullname, {})
        loader = PathLoader(variables, fullname, path, target)

        return Spec(name=fullname, loader=loader)


@attr.s
class Spec:
    name = attr.ib()
    loader = attr.ib()
    submodule_search_locations = attr.ib(factory=list)
    has_location = attr.ib(default=True)
    cached = attr.ib(default=False)
    origin = attr.ib(default=None)


def ancestors(items):
    for i, item in enumerate(reversed(items), start=1):
        yield item, items[:-i]


def get_package_map(strings):
    m = collections.defaultdict(list)
    for string in strings:
        items = string.split(".")
        for module_name, parents in ancestors(items):
            for _parent in parents:
                package_name = ".".join(parents)
                module_full_name = package_name + "." + module_name
                if module_full_name not in m[package_name]:
                    m[package_name].append(module_full_name)

    return dict(m)


def make_finder(mapping):
    package_map = get_package_map(mapping.keys())
    allowed_names = []
    for lst in package_map.values():
        for sub in lst:
            allowed_names.append(sub)

    allowed_names += list(package_map.keys())
    return PathFinder(allowed_names, mapping)



from __future__ import annotations

import re
import textwrap
import types
import unicodedata

from typing import Tuple
from typing import List
from typing import Iterable


import lark


from voca import log
from voca import utils


class Transformer(lark.Transformer):
    def chord(self, arg):
        return arg[0]

    def key(self, arg):

        return arg[0]

    text = list


@log.log_call
def extract(tree: lark.Tree) -> Tuple[str, List]:
    tree = Transformer().transform(tree)
    return tree.data, tree.children


@log.log_call
def extract_commands(tree: lark.Tree) -> Tuple[str, List]:
    tree = Transformer().transform(tree)

    return tree.children


def _replace_match(match):
    body = unicodedata.name(match.group(0)).replace(" ", "_")
    return f"_{body}_"


@log.log_call
def normalize_pattern(text):

    return re.sub(r"\W", _replace_match, text).lower()


@log.log_call
def build_rules(registry: utils.Registry) -> List[utils.Rule]:
    rules = []
    for i, (pattern, function) in enumerate(registry.pattern_to_function.items()):
        normalized_pattern = normalize_pattern(pattern)
        name = f"rule_{i}__{normalized_pattern}"
        rules.append(utils.Rule(name=name, pattern=pattern, function=function))

    return rules


@log.log_call
def build_grammar(registry: utils.Registry, rules: List[utils.Rule]) -> str:

    start = "?start : message_group"

    message_group = "message_group : message+"

    message = "?message : " + "|".join(rule.name for rule in rules)

    rules_segment = "\n".join([f"{rule.name} : {rule.pattern}" for rule in rules])
    patterns_segment = "\n".join([f"{k} : {v}" for k, v in registry.patterns.items()])
    body = "\n".join([message_group, message, rules_segment, patterns_segment])

    imports = textwrap.dedent(
        """\


        %import common.ESCAPED_STRING
        %import common.SIGNED_NUMBER
        %import common.WS
        %import common.CNAME -> NAME
        %import common.NUMBER
        %import common.WS_INLINE

        %ignore WS

        """
    )
    return "\n".join([start, body, imports])


@log.log_call
def combine_modules(modules: Iterable[utils.PluginModule]):

    wrapper_group = utils.WrapperGroup()
    for module in modules:
        wrapper_group.wrappers.append(module.wrapper)

    return wrapper_group


def combine_registries(registries):
    combined = utils.Registry()
    for registry in registries:
        combined.pattern_to_function.update(registry.pattern_to_function)
        combined.patterns.update(registry.patterns)
    return combined



import json
import sys
import types
import os

import pytest

from voca import patching
from voca import caster_adapter

from tests import test_voca
from tests import helpers


@pytest.mark.parametrize(
    "spec,expected",
    [
        ("baz", '"baz"'),
        ("<foo>", "foo"),
        ("[foo]", '[ "foo" ]'),
        ("[<foo>]", "[foo]"),
        ("(foo|bar)", '( "foo" | "bar" )'),
        ("quux <narg>", '"quux" narg'),
        ("quux [<narg>]", '"quux" [narg]'),
        (
            "[( foo bar | baz | eggs spam )] ham <narg>",
            '[( "foo bar" | "baz" | "eggs spam" )] "ham" narg',
        ),
    ],
)
def test_convert_spec(spec, expected):
    got = caster_adapter.convert_spec(spec)
    print(repr(spec), repr(got), repr(expected))
    assert got == expected


@pytest.mark.usefixtures("virtual_display", "config_dir")
def test_simple_caster():
    from voca import caster_adapter

    utterances = ["simple"]

    rows = [test_voca.make_command(utterance, final=True) for utterance in utterances]
    lines = ("\n".join(json.dumps(row) for row in rows) + "\n").encode()

    with helpers.capture_keypresses() as typed:
        helpers.run(
            ["worker", "-i", "voca.plugins.basic", "-i", "voca.plugins.vscode"],
            input=lines,
            env={"VOCA_PATCH_CASTER": "1", **os.environ},
        )

    expected = ["KEY_X"]
    assert typed == expected


@pytest.mark.usefixtures("virtual_display", "hash_seed", "config_dir")
def test_caster_extras():
    from voca import caster_adapter

    utterances = ["scroll page up five"]

    rows = [test_voca.make_command(utterance, final=True) for utterance in utterances]
    lines = ("\n".join(json.dumps(row) for row in rows) + "\n").encode()

    with helpers.capture_keypresses() as typed:
        helpers.run(
            ["worker", "-i", "voca.plugins.basic", "-i", "voca.plugins.vscode"],
            input=lines,
            env={"VOCA_PATCH_CASTER": "1", **os.environ},
        )

    expected = ["KEY_LEFT_ALT", "KEY_PAGE_UP"] * 5
    assert typed == expected


@pytest.mark.usefixtures("virtual_display", "config_dir")
def test_patch_caster():

    namespace = types.SimpleNamespace(python=types.SimpleNamespace(x=1))

    mapping = {
        "castervoice.lib.utilities": {},
        "castervoice.lib.ccr": {},
        "castervoice.lib.ccr.python": {"python": namespace},
        "castervoice.lib.ccr.python.python": {
            "PythonNon": types.SimpleNamespace(mapping={"with": 1})
        },
    }
    finder = patching.make_finder(mapping)

    with patching.finder_patch(finder):
        # import castervoice.lib.ccr
        import castervoice.lib.ccr.python
        import castervoice.lib.ccr.python.python

        assert castervoice.lib.ccr.python.python.PythonNon.mapping["with"] == 1


@pytest.mark.usefixtures("virtual_display", "config_dir")
def test_patch_dragonfly():

    mapping = {"dragonfly": {"Repeat": 5}, "dragonfly.__init__": {}}
    finder = patching.make_finder(mapping)

    with patching.finder_patch(finder):
        from dragonfly import Repeat

        assert Repeat == 5


@pytest.mark.usefixtures("virtual_display", "config_dir")
def test_f_keys():

    from voca import caster_adapter

    utterances = ["go to definition"]

    rows = [test_voca.make_command(utterance, final=True) for utterance in utterances]
    lines = ("\n".join(json.dumps(row) for row in rows) + "\n").encode()

    with helpers.capture_keypresses() as typed:
        helpers.run(
            ["worker", "-i", "voca.plugins.basic", "-i", "voca.plugins.vscode"],
            input=lines,
            env={"VOCA_PATCH_CASTER": "1", **os.environ},
        )

    expected = ["KEY_F12"]
    assert typed == expected


@pytest.mark.usefixtures("virtual_display", "config_dir")
def test_more_keys():

    from voca import caster_adapter

    utterances = ["up", "down", "left", "right"]
    expected = ["KEY_UP", "KEY_DOWN", "KEY_LEFT", "KEY_RIGHT"]
    rows = [test_voca.make_command(utterance, final=True) for utterance in utterances]
    lines = ("\n".join(json.dumps(row) for row in rows) + "\n").encode()

    with helpers.capture_keypresses() as typed:
        helpers.run(
            ["manage", "-i", "voca.plugins.basic", "-i", "castervoice.apps.vscode"],
            input=lines,
            env={"VOCA_PATCH_CASTER": "1", **os.environ},
        )

    assert typed == expected


@pytest.mark.usefixtures("virtual_display", "config_dir")
def test_using_castervoice_apps():

    from voca import caster_adapter

    utterances = ["incremental reverse"]

    rows = [test_voca.make_command(utterance, final=True) for utterance in utterances]
    lines = ("\n".join(json.dumps(row) for row in rows) + "\n").encode()

    with helpers.capture_keypresses() as typed:
        helpers.run(
            [
                "worker",
                "-i",
                "voca.plugins.basic",
                "-i",
                "castervoice.apps.emacs",
                "--no-backup-modules",
            ],
            input=lines,
            env={"VOCA_PATCH_CASTER": "1", **os.environ},
        )

    expected = ["KEY_LEFT_CTRL", "KEY_R"]
    assert typed == expected



"""
Module that contains the command line app.

Why does this file exist, and why not put this in __main__?

  You might be tempted to import things from __main__ later, but that will cause
  problems: the code will get executed twice:

  - When you run `python -mvoca` python will execute
    ``__main__.py`` as a script. That means there won't be any
    ``voca.__main__`` in ``sys.modules``.
  - When you import __main__ it will get executed again (as a module) because
    there's no ``voca.__main__`` in ``sys.modules``.

  Also see (1) from http://click.pocoo.org/5/setuptools/#setuptools-integration
"""
import sys
import types
import os

import click
import eliot

from voca import app
from voca import listen
from voca import manager
from voca import worker
from voca import log


CONTEXT_SETTINGS = {"auto_envvar_prefix": "VOCA"}


def log_cli_call(f):
    def wrap(*args, **kwargs):
        with eliot.start_action(
            action_type="cli_command", pid=os.getpid(), argv=sys.argv
        ):
            return log.log_call(f)(*args, **kwargs)

    return wrap


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option("--log/--no-log", "should_log", is_flag=True, default=True)
@click.pass_context
def cli(ctx, **kwargs):
    ctx.obj = types.SimpleNamespace()
    ctx.obj.kwargs = kwargs.copy()
    app.main(**kwargs)


@cli.command("mic")
def _detect_mic(**kwargs):
    del sys.argv[1]
    import voca.mic


@cli.command("listen")
@click.option("--server", "-s", default="127.0.0.1")
@click.option("--port", "-p", default=8019, type=int)
@click.option("--device", "-d", default=7, type=int)
@click.option("--hypotheses", type=bool, default=True)
@click.option("--save-adaptation-state", type=bool)
@click.option("--send-adaptation-state", type=bool)
@click.option("--audio-gate", default=0, type=int)
@log_cli_call
def _listen(**kwargs):
    content_type = "audio/x-raw, layout=(string)interleaved, rate=(int)16000, format=(string)S16LE, channels=(int)1"
    path = "client/ws/speech"

    listen.run(
        args=types.SimpleNamespace(**kwargs), path=path, content_type=content_type
    )


@cli.command("manage")
@click.option("--import-path", "-i", "module_names", multiple=True, default=None)
@click.option("--num-workers", type=int, default=5)
@click.pass_obj
@log_cli_call
def _manage(obj, **kwargs):
    manager.main(**obj.kwargs, **kwargs)


@cli.command("worker")
@click.option("import_paths", "-i", multiple=True)
@click.option("--patch-caster", is_flag=True, default=False, envvar="VOCA_PATCH_CASTER")
@click.option(
    "--backup-modules/--no-backup-modules",
    "use_backup_modules",
    is_flag=True,
    default=True,
)
@click.pass_obj
@log_cli_call
def _worker(obj, patch_caster, **kwargs):
    if patch_caster:
        from voca import caster_adapter

        caster_adapter.patch_all()
    worker.main(**kwargs)



import os
import sys

import eliot

import voca.log


def main(should_log):

    if not should_log:
        return

    eliot.register_exception_extractor(Exception, voca.log.summarize_exception)
    eliot.add_destinations(voca.log.json_to_file(sys.stdout))



from voca.caster_adapter import R
from voca.caster_adapter import Text
from voca.caster_adapter import Key
from voca.caster_adapter import add_to_registry


from voca import utils

registry = utils.Registry()


mapping = {
    "with": R(Text("with "), rdescript="Python: With"),
    "open file": R(Text("open('filename','r') as f:"), rdescript="Python: Open File"),
    "read lines": R(Text("content = f.readlines()"), rdescript="Python: Read Lines"),
    "try catch": R(
        Text("try:")
        + Key("enter:2/10, backspace")
        + Text("except Exception:")
        + Key("enter"),
        rdescript="Python: Try Catch",
    ),
    "name": R(Text('if __name__ == "__main__":\n\t'), rdescript="Python: name==main"),
}


registry = add_to_registry(mapping, registry)

wrapper = utils.Wrapper(registry)



import functools
import sys
import time

from typing import List
from typing import Callable
from typing import Coroutine
from typing import Awaitable


import trio
import pyautogui
import pynput
import pynput.keyboard

from voca import utils
from voca import platforms
from voca import log

registry = utils.Registry()

registry.define(
    {
        "?any_text": r"/\w.+/",
        "key": utils.regex("|".join(utils.pronunciation_to_value().keys())),
        "chord": 'key ("+" chord)*',
    }
)


@log.log_call
def type_chord(chord):
    keyboard = pynput.keyboard.Controller()
    modifiers = [getattr(pynput.keyboard.Key, mod.name) for mod in chord.modifiers]

    try:
        key = pynput.keyboard.Key[chord.name]
    except KeyError:
        key = chord.name

    if modifiers:
        with keyboard.pressed(*modifiers):
            # XXX This seems like it wouldn't work, but it is working in tests.
            keyboard.press(key)
            keyboard.release(key)
            pyautogui.press(chord.name)

    else:
        # I'm not sure why pynput wasn't working on simple keys, but this seems to work.
        pyautogui.press(chord.name)


@log.log_async_call
async def press(chord: str):
    if isinstance(chord, str):
        await trio.run_sync_in_worker_thread(
            functools.partial(pyautogui.typewrite, [chord])
        )
        return

    await trio.run_sync_in_worker_thread(type_chord, chord)


@log.log_async_call
async def write(message: str):
    await trio.run_sync_in_worker_thread(
        functools.partial(pyautogui.typewrite, message)
    )


@registry.register('"alert" any_text')
@log.log_async_call
async def _alert(text):
    await trio.run_sync_in_worker_thread(functools.partial(pyautogui.alert, text))


@log.log_async_call
async def speak(message):
    await utils.run_subprocess(["say", message])


@registry.register("chord")
@registry.register('"say" chord')
@log.log_async_call
async def _say(message: List[str]):
    [chord_string] = message
    chord_value = utils.pronunciation_to_value()[chord_string]
    await press(chord_value)


@registry.register('"switch" chord')
@log.log_async_call
async def _switch(message: List[str]):
    [chord_string] = message
    chord_value = utils.pronunciation_to_value()[chord_string]
    await press(f"super+{chord_value}")


press_key: Callable = utils.async_runner(press)


registry.pattern_to_function['"monitor"'] = press_key("M")
registry.pattern_to_function['"mouse"'] = press_key("O")


@registry.register('"div0"')
async def _div0(*args):
    1 / 0


wrapper = utils.Wrapper(registry)



from voca import utils
from voca import context
from voca.plugins import basic


registry = utils.Registry()
registry.define({"?text": r"/\w.+/"})


wrapper = utils.Wrapper(registry, context=context.NeverContext())


@registry.register('"nope"')
async def _say_xyz(message):
    await basic.write("xyz")


wrapper = utils.Wrapper(registry, context=context.NeverContext())



import importlib
import functools
import sys
import os
import textwrap
import json
import types
import shutil
import pathlib
import importlib.util

from typing import Iterable
from typing import List
from typing import Tuple

import eliot
import trio
import toml
import lark


from voca import utils
from voca import streaming
from voca import log
from voca import parsing
from voca import context
from voca import config


@log.log_async_call
async def handle_message(wrapper_group: utils.WrapperGroup, data: dict):
    message = data["result"]["hypotheses"][0]["transcript"]

    with eliot.start_action(action_type="parse_command") as action:

        handler = await make_specific_handler(wrapper_group, data)
        tree = handler.parser.parse(message)

    commands = parsing.extract_commands(tree)

    for command in commands:
        rule_name, args = command.data, command.children
        function = handler.rule_name_to_function[rule_name]
        with eliot.start_action(
            action_type="run_command", command=rule_name, args=args, function=function
        ):
            await function(args)


@log.log_call
def load_from_path(import_path, filename):
    spec = importlib.util.spec_from_file_location(import_path, filename)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


@log.log_call
def get_backup_module(import_path, backup_dir):
    sys.path.insert(0, str(backup_dir))
    try:
        module = importlib.import_module(import_path)
    except Exception:
        module = None
    finally:
        del sys.path[0]
    return module


@log.log_call
def save_backup_module(module, import_path, backup_dir):
    new_path = (backup_dir / import_path.replace(".", "/")).with_suffix(".py")
    new_path.parent.mkdir(exist_ok=True, parents=True)
    shutil.copy2(module.__file__, new_path)


@log.log_call
def get_module(import_path, backup_dir, use_backup_modules):
    """Import module and cache it in backup_dir, returning backup on failure."""
    try:
        with eliot.start_action(
            action_type="import_module",
            import_path=import_path,
            sys_path=sys.path,
            sys_meta_path=sys.meta_path,
        ):
            module = importlib.import_module(import_path)
    except Exception:
        if not use_backup_modules:
            raise
        module = get_backup_module(import_path, backup_dir)
    else:
        save_backup_module(module, import_path, backup_dir)
    return module


@log.log_call
def collect_modules(
    import_paths: Iterable[str], use_backup_modules: bool
) -> List[types.ModuleType]:
    backup_dir = pathlib.Path(config.get_config_dir()) / "backup_modules"

    modules = []
    for import_path in import_paths:
        module = get_module(import_path, backup_dir, use_backup_modules)
        if module is not None:
            modules.append(module)
    return modules


def combine_registries(registries):
    combined = utils.Registry()
    for registry in registries:
        combined.pattern_to_function.update(registry.pattern_to_function)
        combined.patterns.update(registry.patterns)
    return combined


async def make_specific_handler(wrapper_group, data):
    filtered = await context.filter_wrappers(wrapper_group, data)
    registry = combine_registries([wrapper.registry for wrapper in filtered.wrappers])

    rules = parsing.build_rules(registry)
    grammar = parsing.build_grammar(registry, rules)
    rule_name_to_function = {rule.name: rule.function for rule in rules}

    parser = lark.Lark(
        grammar, debug=True, lexer="dynamic_complete", maybe_placeholders=True
    )

    return utils.Handler(
        registry=registry, parser=parser, rule_name_to_function=rule_name_to_function
    )


@log.log_async_call
async def async_main(wrapper_group):
    stream = trio._unix_pipes.PipeReceiveStream(os.dup(0))
    receiver = streaming.TerminatedFrameReceiver(stream, b"\n")

    async for message_bytes in receiver:
        data = json.loads(message_bytes.decode())

        try:
            with eliot.Action.continue_task(
                task_id=data.get("eliot_task_id", "@")
            ) as action:
                await handle_message(wrapper_group=wrapper_group, data=data)
        except Exception as e:
            action.finish(e)
            raise

        sys.exit(0)


@log.log_call
def main(import_paths: Tuple[str], use_backup_modules: bool):

    modules = collect_modules(import_paths, use_backup_modules)
    modules = [utils.transform_module(module) for module in modules]

    wrapper_group = parsing.combine_modules(modules)

    trio.run(functools.partial(async_main, wrapper_group=wrapper_group))



from voca import utils
from voca import context
from voca.plugins import basic


registry = utils.Registry()
registry.define({"?text": r"/\w.+/"})


@registry.register('"check"')
async def _say_def(message):
    await basic.write("turt")


wrapper = utils.Wrapper(
    registry, context=context.WindowContext("Python Turtle Graphics")
)



from voca import utils
from voca import context
from voca.plugins import basic

registry = utils.Registry()


registry.define({"?text": r"/\w.+/"})
registry.register("declare text")(basic._alert)


wrapper = utils.Wrapper(
    registry=registry, context=context.WindowContext(title="terminator")
)



from __future__ import annotations

import collections
import contextlib
import functools
import json
import sys
import inspect
import os
import io
import traceback
import types

from typing import Optional
from typing import Callable
from typing import Any
from typing import Union
from typing import Iterable
from typing import Container

import attr
import eliot
import six


@functools.singledispatch
def to_serializable(obj: Any) -> Union[str, list, dict, int, float]:
    try:
        return {"type": type(obj).__name__, "attributes": attr.asdict(obj)}
    except attr.exceptions.NotAnAttrsClassError:
        return str(obj)


@to_serializable.register(types.SimpleNamespace)
def _(obj):
    return vars(obj)


def json_to_file(file: Optional[io.TextIOWrapper] = None) -> Callable:

    if file is None:
        file = sys.stdout

    def _json_to_file(x):
        print(json.dumps(x, default=to_serializable), file=file)

    return _json_to_file


def _exception_lines(exc: BaseException):
    return traceback.format_exception(type(exc), exc, exc.__traceback__)


def _extract_traceback(exc: BaseException):
    return "".join(_exception_lines(exc))


def _exception_data(exc: BaseException):
    exclude = set(dir(Exception())) - {"args", "__cause__", "__context__"}
    return {k: v for k, v in inspect.getmembers(exc) if k not in exclude}


def summarize_exception(exc: BaseException):
    return {
        "traceback": _extract_traceback(exc),
        "exception_data": _exception_data(exc),
    }


def log_call(
    wrapped_function: Optional[Callable] = None,
    action_type: Optional[str] = None,
    include_args: Optional[Iterable[str]] = None,
    include_result: bool = True,
) -> Callable:
    """Decorator/decorator factory that logs inputs and the return result.
    If used with inputs (i.e. as a decorator factory), it accepts the following
    parameters:
    @param action_type: The action type to use.  If not given the function name
        will be used.
    @param include_args: If given, should be a list of strings, the arguments to log.
    @param include_result: True by default. If False, the return result isn't logged.
    """
    if wrapped_function is None:
        return functools.partial(
            log_call,
            action_type=action_type,
            include_args=include_args,
            include_result=include_result,
        )

    if action_type is None:
        action_type = "{}.{}".format(
            wrapped_function.__module__, wrapped_function.__qualname__
        )

    if include_args is not None:

        sig = inspect.signature(wrapped_function)
        if set(include_args) - set(sig.parameters):
            raise ValueError(
                (
                    "include_args ({}) lists arguments not in the " "wrapped function"
                ).format(include_args)
            )

    @functools.wraps(wrapped_function)
    def logging_wrapper(*args, **kwargs):
        bound_args = inspect.signature(wrapped_function).bind(*args, **kwargs)

        # Remove self if it's included:
        if "self" in bound_args.arguments:
            bound_args.arguments.pop("self")

        # Filter arguments to log, if necessary:
        if include_args is not None:
            bound_args.arguments = {k: bound_args.arguments[k] for k in include_args}

        with eliot.start_action(
            action_type=action_type, **bound_args.arguments
        ) as action:

            result = wrapped_function(*args, **kwargs)
            if not include_result:
                return result
            action.add_success_fields(result=result)
            return result

    return logging_wrapper


def log_async_call(
    wrapped_function: Optional[Callable] = None,
    action_type: Optional[str] = None,
    include_args: Optional[Iterable[str]] = None,
    include_result: bool = True,
) -> Callable:
    """Decorator/decorator factory that logs inputs and the return result.
    If used with inputs (i.e. as a decorator factory), it accepts the following
    parameters:
    @param action_type: The action type to use.  If not given the function name
        will be used.
    @param include_args: If given, should be a list of strings, the arguments to log.
    @param include_result: True by default. If False, the return result isn't logged.
    """
    if wrapped_function is None:
        return functools.partial(
            log_call,
            action_type=action_type,
            include_args=include_args,
            include_result=include_result,
        )

    if action_type is None:

        action_type = "{}.{}".format(
            wrapped_function.__module__, wrapped_function.__qualname__
        )

    if include_args is not None:

        sig = inspect.signature(wrapped_function)
        if set(include_args) - set(sig.parameters):
            raise ValueError(
                f"include_args ({include_args}) lists arguments not in the "
                "wrapped function"
            )

    @functools.wraps(wrapped_function)
    async def logging_wrapper(*args, **kwargs):
        bound_args = inspect.signature(wrapped_function).bind(*args, **kwargs)

        # Remove self if it's included:
        if "self" in bound_args.arguments:
            bound_args.arguments.pop("self")

        # Filter arguments to log, if necessary:
        if include_args is not None:
            bound_args.arguments = {k: bound_args.arguments[k] for k in include_args}

        with eliot.start_action(
            action_type=action_type, **bound_args.arguments
        ) as action:

            result = await wrapped_function(*args, **kwargs)

            if include_result:
                action.add_success_fields(result=result)

            return result

    return logging_wrapper



import enum
import platform

_name_to_function = {}


class System(enum.Enum):
    LINUX = enum.auto()
    WINDOWS = enum.auto()
    DARWIN = enum.auto()


def implementation(*which):
    def wrap(func):

        try:
            seen_function = _name_to_function[func.__name__]
        except KeyError:
            seen_function = func
            _name_to_function[func.__name__] = func

        func.implementations = getattr(seen_function, "implementations", {})

        for item in which:
            seen_function.implementations[item] = func

        system_enum_value = System[(platform.system().upper())]

        try:
            return seen_function.implementations[system_enum_value]
        except KeyError:
            return seen_function

    return wrap



import re
import functools
import subprocess
import types
import importlib


from typing import Any
from typing import List
from typing import Dict
from typing import Optional
from typing import Callable
from typing import MutableMapping
from typing import Mapping
from typing import Awaitable

from typing_extensions import Protocol

import attr
import toml
import trio
import lark
import pkg_resources
import importlib_resources


import voca
from voca import context


@attr.s
class Registry:
    pattern_to_function: MutableMapping[str, Callable] = attr.ib(factory=dict)
    patterns: MutableMapping = attr.ib(factory=dict)

    def register(self, pattern: str) -> Callable:
        def _register(function: Callable) -> Callable:
            self.pattern_to_function[pattern] = function
            return function

        return _register

    def define(self, patterns: Optional[Mapping[str, str]] = None, **kwargs):
        if patterns:
            self.patterns.update(patterns)
        self.patterns.update(kwargs)


def pronunciation_to_value() -> MutableMapping[str, Any]:
    text = importlib_resources.read_text(voca, "pronunciation.toml")
    return toml.loads(text)


def value_to_pronunciation():
    return {v: k for k, v in pronunciation_to_value().items()}


async def run_subprocess(
    command: List[str], *, input=None, capture_output=False, **options
):
    if input is not None:
        options["stdin"] = subprocess.PIPE
    if capture_output:
        options["stdout"] = options["stderr"] = subprocess.PIPE

    stdout_chunks: List = []
    stderr_chunks: List = []

    async with trio.Process(command, **options) as proc:

        async def feed_input():
            async with proc.stdin:
                if input:
                    try:
                        await proc.stdin.send_all(input)
                    except trio.BrokenResourceError:
                        pass

        async def read_output(stream, chunks):
            async with stream:
                while True:
                    chunk = await stream.receive_some(32768)
                    if not chunk:
                        break
                    chunks.append(chunk)

        async with trio.open_nursery() as nursery:
            if proc.stdin is not None:
                nursery.start_soon(feed_input)
            if proc.stdout is not None:
                nursery.start_soon(read_output, proc.stdout, stdout_chunks)
            if proc.stderr is not None:
                nursery.start_soon(read_output, proc.stderr, stderr_chunks)
            await proc.wait()

    stdout = b"".join(stdout_chunks) if proc.stdout is not None else None
    stderr = b"".join(stderr_chunks) if proc.stderr is not None else None

    if proc.returncode:
        raise subprocess.CalledProcessError(
            proc.returncode, proc.args, output=stdout, stderr=stderr
        )
    return subprocess.CompletedProcess(proc.args, proc.returncode, stdout, stderr)


def quote(word: str) -> str:
    if word.startswith('"'):
        return f"'{word}'"
    return f'"{word}"'


def regex(word):
    return f"/{word}/"


@attr.dataclass
class Handler:
    registry: Registry
    parser: lark.Lark
    rule_name_to_function: dict


@attr.dataclass
class HandlerGroup:
    handlers: List[Handler]

    async def pick_handler(self, data):
        return self.handlers[0]


class Context(Protocol):
    async def check(self, data=None) -> bool:
        ...


@attr.dataclass
class Wrapper:
    registry: Registry
    context: Context = attr.ib(default=context.AlwaysContext)


@attr.s
class WrapperGroup:
    wrappers: List[Wrapper] = attr.ib(factory=list)


@attr.dataclass
class Rule:
    name: str
    pattern: str
    function: Callable


@attr.dataclass
class PluginModule(Protocol):
    wrapper: Wrapper


@attr.dataclass
class KeyModifier:
    name: str


@attr.dataclass
class SimpleKey:
    name: str


@attr.dataclass
class KeyChord:
    modifiers: List[KeyModifier]
    name: str


def async_runner(async_function: Callable):
    def build(*args, **kwargs) -> Callable:
        @functools.wraps(async_function)
        async def run(_message: str):
            await async_function(*args, **kwargs)

        return run

    return build


def replace(message: str) -> str:
    lookup = pronunciation_to_value()
    pattern = re.compile(r"\b(" + "|".join(lookup.keys()) + r")\b")
    return pattern.sub(lambda x: lookup[x.group()], message)


def plugin_module_paths() -> List[str]:
    return [
        entry_point.module_name + "." + entry_point.name
        for entry_point in pkg_resources.iter_entry_points("voca_plugins")
    ]


MODULE_TRANSFORMERS = []


def transform_module(module):
    for transform in MODULE_TRANSFORMERS:
        module = transform(module)

    return module
